aUF = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceHook-2.1", "AceDB-2.0", "AceConsole-2.0")
aUF.Layouts = {}

local print = function(msg) if msg then DEFAULT_CHAT_FRAME:AddMessage(msg) end end

------------
-- SYSTEM --
------------

function aUF:OnInitialize()
    self.dewdrop = AceLibrary("Dewdrop-2.0")
	self:RegisterDB("aUFDB")
	self:RegisterDefaults('profile', aUF_DEFAULT_OPTIONS)
	
	if self.db.profile.Positions.player then
        self.db.profile.Positions = {}
        for k,v in pairs(aUF_DEFAULT_OPTIONS.Positions) do
            self.db.profile.Positions[k] = v
        end
    end
    
	self:SetupVariables()
	self:UpdateBlizzVisibility()
	self:InitMenu()
	self:RegisterEvents()
	self:LoadStringFormats()
end

function aUF:OnEnable()
	self:SecureHook("SecurePartyHeader_Update","HeaderUpdate")
	self:SecureHook("SecureRaidGroupHeader_Update","HeaderUpdate")	
	aUF:ScheduleRepeatingEvent("agUF_auraPoolSchedule",self.AuraPool, 0.05, self)
	
	aUF:CreateManualUnitBC("player","Player")
	aUF:CreateManualUnitBC("pet","Pet")
	aUF:CreateManualUnitBC("target","Combo")
	aUF:CreateManualUnitBC("targettarget","Metro")
	aUF:CreateManualUnitBC("targettargettarget","Metro")
	aUF:CreateManualUnitBC("focus","Focus")
		
	for k,v in pairs(self.db.profile.subgroups) do
		if v.Exists == true then
			aUF.subgroups[k] = aUF.classes.aUFgroup:new(k,v.Type)
		end
	end
    
    if self.db.profile.pethack then
        aUF:PetHack()
    end
end

function aUF:PetHack()
    for i=1,4 do
        aUF:CreateManualUnitBC("partypet"..i,"Vanilla")  
    end
    self.pethack = true
end

function aUF:HeaderUpdate(frame)
	if (frame.object and frame.object.UpdateTitle) then
		frame.object:UpdateTitle()
	end
end

function aUF:Reset()
	self:ResetDB("profile")
	self:LoadStringFormats()
	self:CallUnitMethods("Reset")
end

function aUF:OnProfileEnable()
	self:CallUnitMethods("Reset")
end

--------------------------------------
-- OBJECT CREATION/METHOD EXECUTION --
--------------------------------------

function aUF:CreateManualUnitBC(unit,type,name)
	local frameName = "aUF"..(name or unit)
	self.nextUnit = unit
	if not type then
		type = "Vanilla" 
	end
	local frame = CreateFrame("Button",frameName,UIParent,"AGUnit"..type.."Template")
	self.nextUnit = nil
end

function aUF:CreateObjectBC(frame,unit,type)
	if not type then
		self.units[frame:GetName()] = self.classes.aUFunit:new(frame,unit)
	else
		self.units[frame:GetName()] = self.classes["aUFunit"..type]:new(frame,unit)
	end
end

function aUF:FindObjects(sortBy, object)
	local table = {}
	if not object then object = "units" end
	if self[object] then
		for _,v in pairs(self[object]) do
			if v[sortBy] then
				if not table[v[sortBy]] then
					table[v[sortBy]] = {}
				end
				tinsert(table[v[sortBy]],v)
			end
		end
	end
	return table
end

function aUF:CallUnitMethods(func,arg,find,type,object)
	if not func then return end
	if find and type then
		if aUF:FindObjects(type)[find] then
			for _,unitObject in pairs(aUF:FindObjects(type,object)[find]) do
				if unitObject[func] and unitObject.unit then
					unitObject[func](unitObject,arg)
				end
			end
		end
	else
		if not object then object = "units" end
		if self[object] then
			for _,unitObject in pairs(self[object]) do
				if unitObject[func] and unitObject.unit then
					unitObject[func](unitObject,arg)
				end
			end
		end
	end
end

-- Subgroup creation
-- Iterates through the saved vars to find an available slot for a new subgroup

function aUF:CreateNewGroup()
	local i = 1
	local k = "group"..i
	while (self.db.profile.subgroups["group"..i] and self.db.profile.subgroups["group"..i].Exists == true) do
		i = i + 1
		k = "group"..i
		
	end
	aUF.subgroups[k] = aUF.classes.aUFgroup:new(k,"raid",true)
	
	-- Update dew menu
	self.mainMenu = aUF:CreateMenu()
	self.dewdrop:Refresh(3)	
end

-----------------------
-- EVENT REGISTERING --
-----------------------

function aUF:RegisterEvents()
    self:aUFRegisterEvent("RAID_ROSTER_UPDATE")
	self:aUFRegisterEvent("PLAYER_REGEN_ENABLED")
end

function aUF:RAID_ROSTER_UPDATE()
    local inRaid = self.inRaid
    if GetNumRaidMembers()>0 then
        self.inRaid = true
    else
        self.inRaid = false
    end
    
    if inRaid ~= self.inRaid then
        if InCombatLockdown() then
            self.raidDirty = true
        elseif self.subgroups then
            for k,v in pairs(self.subgroups) do
                v:SetVisibility()
            end  
        end
    end
end

function aUF:PLAYER_REGEN_ENABLED()
    if self.raidDirty and self.subgroups then
        for k,v in pairs(self.subgroups) do
            v:SetVisibility()
        end
        self.raidDirty = nil
    end
end

------------------
-- EVENT SYSTEM --
------------------

function aUF:aUFRegisterEvent(event)
	if not self.coreEvents[event] then
		self.coreEvents[event] = {}
	end
	
	self.coreEvents[event] = method or event
	
	self["pool"..event] = function(self,unit) self:TriggerEvent(event,unit) end
	self:RegisterEvent(event,"pool"..event)
end

function aUF:aUFUnregisterEvent(event,method)	
	self.coreEvents[event] = nil
	
	if self.events[event] then
		if next(self.events[event]) == nil then
			if self:IsEventRegistered(event) then
				self:UnregisterEvent(event)
			end
		end
	end
end

function aUF:ObjectRegisterEvent(object,event,method)
	if not self.events[event] then
		self.events[event] = {}
	end
	
	self.events[event][object] = method or event
	
	self["pool"..event] = function(self,unit) self:TriggerEvent(event,unit) end
	self:RegisterEvent(event,"pool"..event)
end

function aUF:ObjectUnregisterEvent(object,event)
	if self.events[event] then
		self.events[event][object] = nil
		if next(self.events[event]) == nil and not self.coreEvents[event] then
			if self:IsEventRegistered(event) then
				self:UnregisterEvent(event)
			end
		end
	end
end

function aUF:TriggerEvent(event,unit)
    if self.coreEvents[event] and self[self.coreEvents[event]] then
        self[self.coreEvents[event]](self)
    end
    if not self.events[event] then return end
	if unit and (self.unitid[unit]) then
		for object,_ in pairs(self.unitid[unit]) do
			local method = self.events[event][object]
			if (method) then
				object[method](object,unit)
			end
		end
	else
		for object,method in pairs(self.events[event]) do
			object[method](object)
		end	
	end
end

-------------------
-- UNIT EVENTS --
-------------------

function aUF:AuraPool()
	local n = 0
	for k,v in pairs(self.auraUpdatePool) do
		if v == true and self.auraUpdates + n < 5 and self.units[k] then
			self.units[k]:UpdateAuras(true)
			n = n + 1
		else
			break
		end
	end
	aUF.auraUpdates = 0
end

function aUF:FeedbackUpdate()
	local maxalpha = 0.6
	local found
	for objectName,v in pairs(self.feedback) do
		found = true
		local unitOjbect = aUF.units[objectName]
		local elapsedTime = GetTime() - unitOjbect.feedbackStartTime
		if ( elapsedTime < COMBATFEEDBACK_FADEINTIME ) then
			local alpha = maxalpha*(elapsedTime / COMBATFEEDBACK_FADEINTIME)
			unitOjbect.HitIndicator:SetAlpha(alpha)
		elseif ( elapsedTime < (COMBATFEEDBACK_FADEINTIME + COMBATFEEDBACK_HOLDTIME) ) then
			unitOjbect.HitIndicator:SetAlpha(maxalpha)
		elseif ( elapsedTime < (COMBATFEEDBACK_FADEINTIME + COMBATFEEDBACK_HOLDTIME + COMBATFEEDBACK_FADEOUTTIME) ) then
			local alpha = maxalpha - maxalpha*((elapsedTime - COMBATFEEDBACK_HOLDTIME - COMBATFEEDBACK_FADEINTIME) / COMBATFEEDBACK_FADEOUTTIME)
			unitOjbect.HitIndicator:SetAlpha(alpha)
		else
			unitOjbect.HitIndicator:Hide()
			aUF.feedback[objectName] = nil
		end
	end
	if not found then 
		self:CancelScheduledEvent("agUF_CombatSchedule")
	end
end

-----------
-- Units --
-----------

function aUF:RegisterUnit(object, unit)
	-- prepare unitid table
	if not self.unitid then
		self.unitid = {}
	end
	
	-- prepare unitid[unit] storage
	if (not self.unitid[unit]) then
		self.unitid[unit] = {}
	end
	
	-- remove current reference
	aUF:UnregisterUnit(object)
	
	-- register new unit
	self.unitid[unit][object] = true
end

function aUF:UnregisterUnit(object)
	for id,objs in pairs(self.unitid) do
		if (objs[object]) then
			objs[object] = nil
		end
	end
end

-----------
-- HOOKS --
-----------

aUF.hooks = {}

function aUF.hooks.hookShowHide(frame)
	local show, hide = frame.Show, frame.Hide
	frame.Show = function(self)
		if (not self.isVisibleFlag) then
			show(self)
			self.isVisibleFlag = true
		end
	end
	frame.Hide = function(self)
		if (self.isVisibleFlag) then
			hide(self)
			self.isVisibleFlag = false
		end
	end
	frame.WillHide = function(self)
		self.doShow = false
	end
	frame.WillShow = function(self)
		self.doShow = true
	end
	frame.DoShowHide = function(self)
		if (self.doShow) then self:Show() else self:Hide() end
	end
	
	frame:Show()
end

function aUF.hooks.hookSetTexture(frame)
	local settexture = frame.SetTexture
	frame.SetTexture = function(self, texture)
		if (self.lasttexture ~= texture) then
			self.lasttexture = texture
			settexture(self, texture)
		end
	end
end
